%Comando A\b:
%Resuelve el sistema Ax=b  mediante un solver polimórfico basado en librerías LAPACK:
%   a) Si es triangular dispersa emplea sustitución regresiva o progresiva.
%   b) si es cuadrada, simétrica y con diagonal positiva intenta Cholesky.
%   c) si es cuadrada y lo anterior falla, intenta LU con pivote (PALU)
%   d) si la matriz no es cuadrada o es singular (o casi) busca una solución en el sentido de los mínimos cuadrados: \min\| Ax-b\|_2^2 \to A'*A x= A'*b (método SVD)
%
%
A=rand(4,4)
b=ones(4,1)
A\b
%
%
% El comando  / suele ser más rápido que inv(A)*b
A=hilb(10);
b=ones(10,1);
%
tic;
A\b;
toc
%
tic;
inv(A)*b;
toc
%
% Herramientas para discutir un sistema:
Amp=[A,b];
rank(Amp) % proporciona el rango de Amp
rref(Amp) % proporciona la forma reducida de la matriz Amp
%
%
%Ejemplo de sistema lineal mal condicionado
m=10;   % Cambiar la dimensión de esta matriz y observar resultados
A=triu(ones(m))-2*eye(m); 
b=zeros(m,1); 
b(end)=10^(-8); 
norm(A\b-zeros(m,1))
%Herramientas para medir los posibles errores de redondeo.
%cond(A,p)  Condicionamiento de A en norma p
cond(A,2)
help cond
% norm(A,p)  Norma de A en norma p
norm(A,2)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Aplicación 1:   la mejor aproximación por mínimos cuadrados discreta a datos (x_i, y_i) 
% de grado menor o igual a m viene dado por:
%
m=1;
x=1:7;
y=[11,15,20,23,26,27,30];
A=vander (x,m+1)
p=(A\y')'
polyout(p,'t')
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Aplicación 2: resolución numérica de problemas de valores en la frontera.
%
%
% difusion.m -- Calcula una aproximación por diferencias 
% finitas centradas del problema 
% de contorno
% \nu u_{xx} -k u= 0 , x\in[a,b],  u(a)= \alpha, u (b)= \beta 
% Este modelo correspondería a la solucion estacionaria de % un proceso difusivo en el que simultaneamente hay una 
% degradación/mortandad.
%
% datos del problema
a = 0;     
b = 2; 
alpha = 1;
beta = 2;
nu = 0.1;   
k = 0.5;
% datos de la discretización
m = 100;                   % numero de nodos interiores
h = (b-a)/(m+1); 
% Construcción del SEL
e = ones(m,1);
T = spdiags([e -2*e e],[-1 0 1],m,m);			 
T= nu*T/h^2;
I = speye(m);
A= T-k*I;
% versión matrices completas (para jugar)
%ones(m,1); em=ones(m-1,1); 
%T = diag(-2*e,0) +diag(em,1)+diag(em,-1);		
%T= nu*T/h^2;
%I = eye(m);
%A= T-k*I;
% definimos el termino independiente que tiene en cuenta 
% los valores en la frontera.
rhs=zeros(m,1);
rhs(1) = - nu*alpha/h^2;
rhs(m) = - nu*beta/h^2;
% Resolvemos el sistema lineal:
tic;
uvec = A\rhs;  
toc
% Recopilamos la solucion numerica y la representamos
x = linspace(a,b,m+2);   % Puntos del mallado incluidos los % contornos
usoln = [alpha; uvec; beta];
plot(x,usoln)
%
% Obsérvese que Octave permite dar un tratamiento especial a matrices dispersas.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Aplicación 3: programación del método de sustitución regresiva
%
function [x] =sustregr(A,b)
% Esta funcion resuelve un sistema de ecuaciones lineales A x = b con  
% A una matriz triangular superior regular mediante sustitucion regresiva.
%
% Datos necesarios para llamar a la funcion:
%      A, matriz de coeficientes del SEL regular y triangular superior.
%      b, vector de terminos independientes (vector/es columna).
% La funcion devuelve como respuesta:
%      x,  solucion del SEL: A^-1 b .
%
[dim,dim]= size(A);
[nfb,ncb]=size(b);
x=zeros(dim,ncb);
for k = dim:-1:1 
  x(k,:)= (b(k,:) - A(k,:)*x)./A(k,k); 
end
end
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Aplicación 4: programación del método de Gauss con pivote parcial
%
%
function [x,vecperm] =gausspiv(A,b)
% Esta funcion transforma un sistema de ecuaciones lineales A x = b con  
% A una matriz regular en un sistema equivalente T x = c con T una matriz 
% triangular superior, mediante eliminacion gaussiana con pivoteo parcial.
% Posteriormente este nuevo sistema se resuelve mediante sustitucion regresiva.
%
% Datos necesarios para llamar a la funcion:
%      A, matriz de coeficientes del SEL regular.
%      b, vector de terminos independientes (vector/es columna).
% La funcion devuelve como respuesta:
%      x, solucion del SEL.
%      vecperm, vector indicando las permutaciones realizadas en el proceso.
%
[dim,dim]= size(A);
[nfb,p]=size(b);
Aamp=[A b];
vecperm=1:dim;
for k = 1:dim-1 
  % Pivoteo parcial dentro de la columna k
  [Y,j]=max(abs(Aamp(k:dim,k)));
  l= j+k-1; % fila donde se encuentra el pivote
  if l > k
     Aamp([k,l],:)=Aamp([l,k],:);
     vecperm([k,l])=vecperm([l,k]);
  end
  % Eliminacion bajo el pivote
  for h=k+1:dim
    m=Aamp(h,k)/Aamp(k,k);
    Aamp(h,k:dim+p)=Aamp(h,k:dim+p)-m*Aamp(k,k:dim+p);
  end
end
% sustitucion regresiva
x=sustregr(Aamp(1:dim,1:dim),Aamp(1:dim,dim+1:dim+p));
end % del function

%%%%
% Este comando emplea la sustitución regresiva implementada en un ejemplo anterior.

  